@@ -4,13 +4,15 @@ import pandas as pd
 import datetime as dt
 import itertools as it
 from scipy.ndimage import label
+import dateutil.relativedelta as durel
 
 from domino import agg
 from domino.categorical_analysis import get_transmat, synthetic_states_from_transmat
 from domino.util import holm_bonferroni_correction, split_to_contiguous, is_time_type, make_all_dims_coords, drop_scalar_coords, squeeze_da,offset_time_dim
-from domino.filtering import *
+from domino.filtering import ds_large_regions, convolve_pad_ds
 from domino.deseasonaliser import Agg_Deseasonaliser
 
+
 class LaggedAnalyser(object):
     """Analysis of lagged composites defined with respect to a categorical event series
     
@@ -153,8 +155,8 @@ class LaggedAnalyser(object):
         self._check_offset_is_valid(offset,mode)
         
         #The meat of the function:  BREAKING CHANGE WITH RESPECT TO PREVIOUS VERSION
-        time_offset=dt.timedelta(**{offset_unit:offset})
-        new_dim=pd.to_datetime(self.variables[offset_dim])-time_offset #THIS USED TO BE +time_offset
+        time_offset=durel.relativedelta(**{offset_unit:offset})
+        new_dim=pd.to_datetime(self.variables[offset_dim]).map(lambda t: t- time_offset) #THIS USED TO BE +time_offset
         self._lagged_variables[offset]=self.variables.copy(deep=False)
         self._lagged_variables[offset][offset_dim]=new_dim
         return
@@ -249,8 +251,7 @@ class LaggedAnalyser(object):
         else:
                 comp=self._aggregate_from_ix(cat_ds,ix,dim,cat_func,cat_vals)
         comp.attrs=ds.attrs
-        return comp.assign_coords({'lag':[lag]})
-
+        return comp.assign_coords({'lag':[lag]})    
     
     #loops over all lags, calling _composite_from_ix, and assembles composites into a single dataset
     def _compute_aggregate_over_lags(self,da,dim,lag_vals,con_func,cat_func):
@@ -400,7 +401,8 @@ class LaggedAnalyser(object):
         *synth_mode*
             A string, specifying how synthetic event indices are to be computed. Valid options are:
             "random": categorical values are randomly chosen with the same probability of occurrence as those found in *LaggedAnalyser.event*, but with no autocorrelation.
-            'markov': A first order Markov chain is fitted to *LaggedAnalyser.event*, producing some autocorrelation and state dependence in the synthetic series. Generally a better approximation than "random" and so should normally be used.
+            "markov": A first order Markov chain is fitted to *LaggedAnalyser.event*, producing some autocorrelation and state dependence in the synthetic series. Generally a better approximation than "random" and so should normally be used.
+            "shuffle": The values are randomly reordered. This means that each value will occur exactly the same amount of times as in the original ix, and so is ideal for particularly rare events or short series.
             
         *data_vars*
             An iterable of strings, specifying for which variables bootstraps should be computed.
@@ -440,13 +442,15 @@ class LaggedAnalyser(object):
         if (self.boot_indices is None)|(not reuse_ixs):
             
             ix_vals,ix_probs,L=self._get_bootparams(da)
-            ixs=self._get_synth_indices(bootnum,synth_mode,da,dim,ix_vals,ix_probs,L)
+            ixs=self._get_synth_indices(da.values,bootnum,synth_mode,da,dim,ix_vals,ix_probs,L)
             self.boot_indices=ixs
         else:
             ixs=self.boot_indices
             print('Reusing stored boot_indices, ignoring new boot parameters.')
         
-        boots=[self._composite_from_ix(ix,ds,dim,con_func,cat_func,lag) for ix in ixs]
+        boots=[make_all_dims_coords(\
+                self._composite_from_ix(ix,ds,dim,con_func,cat_func,lag)\
+             ) for ix in ixs]
         return xr.concat(boots,'boot_num')
     
     #Gets some necessary variables
@@ -455,7 +459,7 @@ class LaggedAnalyser(object):
         return ix_vals,ix_probs/len(da),len(da)
     
     #compute indices
-    def _get_synth_indices(self,bootnum,mode,da,dim,ix_vals,ix_probs,L):
+    def _get_synth_indices(self,index,bootnum,mode,da,dim,ix_vals,ix_probs,L):
         
         ixs=[]
         if mode=='markov':
@@ -467,6 +471,13 @@ class LaggedAnalyser(object):
         elif mode=='random':
             for n in range(bootnum):
                 ixs.append(np.random.choice(ix_vals,size=L,p=list(ix_probs)))
+                
+        elif mode=='shuffle':
+            for n in range(bootnum):
+                ixv=index.copy()
+                np.random.shuffle(ixv)
+                
+                ixs.append(ixv)
         else:
             raise(ValueError(f'synth_mode={synth_mode} is not valid.'))
             
@@ -570,19 +581,28 @@ class LaggedAnalyser(object):
         return self.composite_sigs
     
     
-    def deseasonalise_variables(self,variable_list=None,dim='time'):
+    def deseasonalise_variables(self,variable_list=None,dim='time',agg='dayofyear',smooth=1,coeffs=None):
         if variable_list is None:
             variable_list=list(self.variables)
         for var in variable_list:
             da=self.variables[var]
             dsnlsr=Agg_Deseasonaliser()
-            dsnlsr.fit_cycle(da)
-            cycle=dsnlsr.evaluate_cycle(da[dim])
+            if coeffs is None:
+                dsnlsr.fit_cycle(da,dim=dim,agg=agg)
+            else:
+                dsnslr.cycle_coeffs=coeffs[var]
+                
+            cycle=dsnlsr.evaluate_cycle(data=da[dim],smooth=smooth)
             self.variables[var]=da.copy(data=da.data-cycle.data)
             dsnlsr.data=None #Prevents excess memory storage
             self.deseasonalisers_[var]=dsnlsr
         return
     
+    def get_seasonal_cycle_coeffs(self):
+        
+        coeffs=xr.Dataset({v:dsnlsr.cycle_coeffs for v,dsnlsr in self.deseasonalisers_.items()})
+        return coeffs
+
     #If deseasonalise_variables has been called, then this func can be used to compute the
     #seasonal mean state corresponding to a given composite. This mean state+ the composite
     # produced by self.compute_composites gives the full field composite pattern.
@@ -599,7 +619,8 @@ class LaggedAnalyser(object):
             mean_states[var]=xr.concat([\
                              xr.concat([\
                                     self.deseasonalisers_[var].cycle_coeffs.sel(\
-                                    {agg:getattr(pd.to_datetime(t)+dt.timedelta(**{self.offset_unit:int(l)}),agg)}\
+                                    {agg:getattr(pd.to_datetime(t).map(\
+                                        lambda ti: ti+durel.relativedelta(**{self.offset_unit:int(l)})),agg)}\
                                     ).mean(agg).assign_coords({'lag':l,'index_val':i})\
                             for l in lags],'lag')\
                             for i,t in ts.items()],'index_val')
@@ -620,7 +641,7 @@ class PatternFilter(object):
             self._parse_analyser(analyser)
             
     def __repr__(self):
-        'A PatternFilter object'
+        return 'A PatternFilter object'
         
     def __str__(self):
             return self.__repr__
@@ -649,15 +670,20 @@ class PatternFilter(object):
         self.update_mask(value_mask,mode)
         return
     
-    def apply_area_mask(self,n,dims=None,mode='intersection'):
+    def apply_area_mask(self,n,dims=None,mode='intersection',area_type='gridpoint'):
         
-        
-        area_mask=ds_large_regions(self.mask_ds,n,dims=dims)
+        if area_type=='gridpoint':
+            area_based=False
+        elif area_type=='spherical':
+            area_based=True
+        else:
+            raise(ValueError(f"Unknown area_type {area_type}. Valid options are 'gridpoint' and 'spherical'"))
+        area_mask=ds_large_regions(self.mask_ds,n,dims=dims,area_based=area_type)
         self.update_mask(area_mask,mode)
         return
     
     
-    def apply_convolution(self,n,dims=None,mode='intersection'):
+    def apply_convolution(self,n,dims=None,mode='replace'):
         
         convolution=convolve_pad_ds(self.mask_ds,n,dims=dims)
         self.update_mask(convolution,mode)
@@ -721,7 +747,7 @@ class IndexGenerator(object):
             weights=np.cos(np.deg2rad(ix[lat_dim]))
             return ix.weighted(weights).sum(dims)
             
-    def generate(self,pattern_ds,series_ds,dim='time',ix_means=None,ix_stds=None,slices=None):
+    def generate(self,pattern_ds,series_ds,dim='time',ix_means=None,ix_stds=None,slices=None,drop_blank=False):
         #Parse inputs
         
         if slices is None:
@@ -743,12 +769,19 @@ class IndexGenerator(object):
         
         #Compute indices
         indices=[self._generate_index(pattern_ds,series_ds,dim,sl)\
-                for sl in slices]
+                for sl in self.slices]
         try:
             indices=xr.merge(indices)
         except:
             indices=xr.merge(indices,compat='override')
             print('Warning: merge of indices with "compat=override" required.\n Index dataset may have innacurrate metadata.')
+        #Optionally remove indices which are all nan    
+        if drop_blank:
+            drop=(~indices.isnull()).sum()==0
+            drop=[k for k,d in drop.to_dict()['data_vars'].items() if d['data']]
+            indices=indices.drop_vars(drop)
+            _=[(self.means.pop(x),self.stds.pop(x)) for x in drop]
+            
         return indices
     
     def _generate_index(self,pattern_ds,series_ds,dim,sl):
@@ -782,7 +815,10 @@ class IndexGenerator(object):
         func=self._rename_function
         return index.rename({v:func(v,sl) for v in index.data_vars})
     
-    def get_standardisation_params(self):
-        
-        return self.means,self.stds
+    def get_standardisation_params(self,as_dict=False):
+        if as_dict:
+            return self.means,self.stds
+        else:
+            params=[xr.Dataset(self.means),xr.Dataset(self.stds)]
+            return xr.concat(params,'param').assign_coords({'param':['mean','std']})
         
\ No newline at end of file
